/************************************************************************

 **************************************************************************/

package simple;

import java.util.Collection;
import java.util.Hashtable;
import java.util.List;

import kernel.AbstractAgent;
import kernel.Constraint;
import kernel.ExternalVar;
import kernel.VarIdentifier;
import kernel.constraints.integer.ExtensionBinConstraint;
import kernel.constraints.integer.SAbsolute;
import kernel.constraints.integer.SDistanceXYC;
import kernel.constraints.integer.SEqualXYC;
import kernel.constraints.integer.SGreaterThanEqualXYC;
import kernel.constraints.integer.SGreaterThanXYC;
import kernel.constraints.integer.SLessThanEqualXYC;
import kernel.constraints.integer.SLessThanXYC;
import kernel.constraints.integer.SNotEqualXYC;
import kernel.constraints.integer.SUnaryIntConstraint;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntValSelector;
import kernel.vars.integer.IntVar;
import kernel.vars.integer.IntVariable;
import kernel.vars.integer.MinIntValSelector;
import simulation.metric.Metric;
import util.Logger;
import util.VarAssignment;

/**
 * An abstract agent for the simple mode. This class defines methods to make internal variable and all types of external constraints for a simple
 * model (One variable per agent).
 */
public abstract class SimpleAgent extends AbstractAgent implements SimpleModel {

	/**
	 * The single internal variable controlled by the agent
	 */
	public IntVariable var;
	/**
	 * A value selector for the controlled variable
	 */
	protected IntValSelector valSelector;
	public final Constraint TRUE = null;
	public final Constraint FALSE = null;

	/**
	 * Constructs a simple agent
	 * 
	 * @param agentId
	 *            the ID of the agent
	 * @param agName
	 *            the name of the agent
	 * @param description
	 *            A string given a description the agent
	 */
	public SimpleAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.valSelector = new MinIntValSelector(this);
	}

	public IntVar makeVariable(VarIdentifier varId) {
		if (this.agentId == varId.getAgentId()) return this.var;
		else return this.makeExternalVar(varId);
	}

	/**
	 * {@inheritDoc}
	 */

	public int getNbVars() {
		return 1;
	}

	public int getMaxDomainSize() {
		return this.var.getDomainSize();
	}

	/**
	 * {@inheritDoc}
	 */

	public String solutionToString() {
		return this.var.toString();
	}

	public String toString() {
		return "Agent-" + this.agentId + " var-" + this.var + " : \n";
	}

	// //////////////////////////////////////////////////////////////////////////
	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(int[] domain) {
		this.var = new IntVariable(this, domain);
		this.var.setVarIdentifier(new VarIdentifier(this.agentId, 0));
		this.addVariable(var.getVarIdentifier(), var);
		return var;
	}

	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(int min, int max) {
		this.var = new IntVariable(this, min, max);
		this.var.setVarIdentifier(new VarIdentifier(this.agentId, 0));
		this.addVariable(var.getVarIdentifier(), var);
		return this.var;
	}

	@Override
	public IntExternalVar makeExternalVar(VarIdentifier varId) {
		return this.makeExternalVar(new VarAssignment(varId, Integer.MIN_VALUE));
	}

	public IntExternalVar makeExternalVar(IntVar v) {
		if (!this.containsVariable(v.getVarIdentifier())) {
			IntExternalVar exV = new IntExternalVar(this, v);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(v.getVarIdentifier());
	}

	public IntExternalVar makeExternalVar(VarAssignment v) {
		if (!this.containsVariable(v.getVarIdentifier())) {
			IntExternalVar exV = new IntExternalVar(this, v);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(v.getVarIdentifier());

	}

	public IntExternalVar makeExternalVar(VarIdentifier vId, int min, int max) {
		if (!this.containsVariable(vId)) {
			IntExternalVar exV = new IntExternalVar(this, vId, min, max);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(vId);
	}

	public IntExternalVar makeExternalVar(VarIdentifier vId, int[] dom) {
		if (!this.containsVariable(vId)) {
			IntExternalVar exV = new IntExternalVar(this, vId, dom);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(vId);
	}

	public ExternalVar getExternalVar(int agId) {
		VarIdentifier vId = new VarIdentifier(agId);
		if (this.containsVariable(vId))
			return (ExternalVar) this.getVariable(vId);
		return null;
	}

	// //////////////////////////////////////////////////////////////////////////
	// ///////////// Unary constraints /////////////////////////////////////
	/**
	 * Make an unary constraint that maps each value to a cost
	 */
	public Constraint valuesCosts(Hashtable<Integer, Integer> costsMap) {
		Constraint exconstraint = new SUnaryIntConstraint(var, costsMap);
		this.addConstraint(exconstraint);
		return exconstraint;
	}

	/**
	 * Makes unary constraint X = cste
	 * 
	 * @param cste
	 *            A constant
	 */
	public Constraint eqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.eq, var, cste);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X != cste
	 * 
	 * @param cste
	 *            A constant
	 */

	public Constraint neqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.neq, var, cste);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X <= cste
	 * 
	 * @param cste
	 *            A constant
	 */

	public Constraint leqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.lt, var, cste + 1);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X >= cste
	 * 
	 * @param cste
	 *            A constant
	 */

	public Constraint geqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.gt, var, (cste + 1));
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X < cste with the given cost.
	 * 
	 * @param cste
	 *            A constant
	 */
	public Constraint ltC(int cste) {
		return this.leqC(cste - 1);
	}

	/**
	 * Makes unary constraint X > cste
	 * 
	 * @param cste
	 *            A constant
	 */

	public Constraint gtC(int cste) {
		return this.geqC(cste + 1);
	}

	// //////////////////////////////////////////////////////////////////////////
	/**
	 * Make the constraint X = Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint eqY(IntExternalVar y) {
		Constraint exC = this.eqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X <= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint leqY(IntExternalVar y) {
		Constraint exC = this.leqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X >= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint geqY(IntExternalVar y) {
		Constraint exC = this.geqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X <= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint ltY(IntExternalVar y) {
		Constraint exC = this.leqYplusC(y, -1);
		return exC;
	}

	/**
	 * Make the constraint X >= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint gtY(IntExternalVar y) {
		Constraint exC = this.geqYplusC(y, +1);
		return exC;
	}

	/**
	 * Make the constraint X = |Y| when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint eqAbsY(IntExternalVar y) {
		SConstraint exC = new SAbsolute(this.var, y);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X = |Y + C| when X is the internal variable, Y is the given external variable <code>y</code> and C is a positive constant.
	 */
	public Constraint eqAbsYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SAbsolute(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X != Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint neqY(IntExternalVar y) {
		Constraint exC = this.neqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X = Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint eqYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X != Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint neqYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SNotEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X <= Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint leqYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SLessThanEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X < Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint ltYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SLessThanXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X >= Y + C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a constant
	 */
	public Constraint geqYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SGreaterThanEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X > Y + C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a positive constant.
	 */
	public Constraint gtYplusC(IntExternalVar y, int c) {
		SConstraint exC = new SGreaterThanXYC(this.var, y, c + 1);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint |X - Y| Operator C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a positive
	 * constant.
	 */
	public Constraint distanceXYC(IntExternalVar y, int c, int operator) {
		if (operator == Constraint.GEQ) {
			c--;
			operator = Constraint.GT;
		}
		if (operator == Constraint.LEQ) {
			operator = Constraint.LT;
			c++;
		}
		SConstraint exC = new SDistanceXYC(this.var, y, c, operator);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make an extensional binary constraints with non allowed tuples.
	 */
	public Constraint feasiblePair(IntExternalVar y, List<int[]> mat) {
		Constraint exC = new ExtensionBinConstraint(Constraint.feasP, this.var, y, mat);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make an extensional binary constraints with non allowed tuples.
	 */
	public Constraint infeasiblePair(IntExternalVar y, List<int[]> mat) {
		Constraint exC = new ExtensionBinConstraint(Constraint.infeasP, this.var, y, mat);
		this.addConstraint(exC);
		return exC;
	}

	@Override
	public void logMyDomain() {
		if (Logger.LOG_DEBUG) {
			this.log("My " + this.var.getDomain().toString(), Logger.INFO_LEVEL);
		}
	}

	public Constraint makePairAC(IntExternalVar exv, List<int[]> list, boolean feas) {
		if (feas) { return this.feasiblePair(exv, list); }
		return this.infeasiblePair(exv, list);
	}

	@Override
	public Collection<Metric> getAuxiliaryStatistics() {
		this.auxiliaryStatistics.put("nbAssignments", new Metric("nbAssignments", this.var.getTimeCounter()));
		return this.auxiliaryStatistics.values();
	}
}
