package simple.dcsp.abtfamily.abtdo;

import java.util.HashMap;

import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.StopException;
import kernel.com.Message;
import kernel.com.TerminateMsgHandler;
import kernel.events.AgentEvent;
import simple.dcsp.abtfamily.IdleEvent;
import simple.dcsp.abtfamily.abtdo.ABTDOAgent.ABTDOCommunicator;
import util.Commands;
import util.Instantiation;
import util.IntArray;
import util.Logger;
import util.VarAssignment;

/**
 * termination detection (Silaghi et al. 2000g)
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class ABTDOTermDetection extends TerminateMsgHandler {

	protected IntArray lastReceiver = null;

	protected CurrentOrder bestOrder = null;
	protected HashMap<Integer, Instantiation> acceptedMsg = null;
	protected HashMap<Integer, CurrentOrder> currentOrders = null;
	protected CurrentOrder prevSentOrder = null;
	protected Instantiation prevSentAccepted = null;

	public ABTDOTermDetection(AbstractAgent agent) {
		super(agent);
		this.acceptedMsg 	= new HashMap<Integer, Instantiation>();
		this.currentOrders 	= new HashMap<Integer, CurrentOrder>();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.TerminateMsgHandler#processTerminationDetection(.com.Message)
	 */
	public boolean processTerminationDetection(Message msg) throws StopException {
		Instantiation accepted = (Instantiation) msg.get(Commands.ACCEPTED); 
		CurrentOrder rcvOrder  = (CurrentOrder) msg.get(Commands.ORDER);
		int senderId = msg.getSenderId();
		this.setBestOrder(rcvOrder);
		this.currentOrders.put(senderId, rcvOrder);
		this.acceptedMsg.put(senderId, accepted);
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.TerminateMsgHandler#handleAgentEvent(events.AgentEvent )
	 */
	@Override
	public void handleAgentEvent(AgentEvent agentEvent) {
		if (agentEvent instanceof IdleEvent && !this.myAgent.isFinished()) {
			this.tryToSendAccepted();
		}
		if (agentEvent instanceof kernel.events.AgentFinishedEvent) {
			this.askMasterToTerminate(true);
		}
	}

	/**
	 * 
	 */
	public void tryToSendAccepted() {
		CurrentOrder myCurrentOrder = ((ABTDOAgent) this.myAgent).getCurrentOrder();
		this.myAgent.log("I try to send an accepted msg \n"+
				"myCurrentOrder: "+myCurrentOrder+"\n"+
				"Orders   : "+this.currentOrders+"\n"+
				"Accepted : "+this.acceptedMsg+"\n"+
				"bestOrder: "+this.bestOrder, Logger.INFO_LEVEL);
		Instantiation accepted = this.allAcceptedMsgHaveBeenReceived();
		if (accepted != null) {
			if (this.ISendTheSame(accepted, myCurrentOrder)) {
				this.myAgent.log("I sent the same "+accepted+" ", Logger.INFO_LEVEL);
				return;
			}
			this.prevSentAccepted = accepted.clone();
			this.prevSentOrder = myCurrentOrder.clone();
			IntArray higherNeighbors = this.myAgent.communicator.getHigherNeighbors();
			if (higherNeighbors.size() != 0) { // am I an initial end agent
				for (int higher : higherNeighbors) {
					this.sendAccepted(higher);
				}
			}
			if(((ABTDOCommunicator) this.myAgent.communicator).getMyHigherNeighbors().size()==0){
				this.sendAccepted(AbstractMaster.MasterId);
			}
		}
	}

	private boolean ISendTheSame(Instantiation accepted, CurrentOrder currentOrder) {
		if (this.prevSentAccepted != null && this.prevSentAccepted.equals(accepted) && this.prevSentOrder != null
				&& this.prevSentOrder.equals(currentOrder)) { return true; }
		return false;
	}

	protected void sendAccepted(int receiverId) {
		Message msg = this.createMessage();
		msg.addContents(Commands.ACCEPTED, this.prevSentAccepted.clone());
		msg.addContents(Commands.ORDER, this.prevSentOrder.clone());
		msg.setReceiverId(receiverId);
		this.myAgent.sendMessage(msg);
	}

	protected synchronized Instantiation allAcceptedMsgHaveBeenReceived() {
		IntArray lowerNeighbors = this.myAgent.communicator.getLowerNeighbors();
		Instantiation myView = this.myAgent.agentView.getContextsOfHigherAgents();
		VarAssignment myAssign = ((ABTDOAgent) this.myAgent).var.getVarAssignment();
		myView.add(myAssign);
		for (int lower : lowerNeighbors) {
			if (!this.currentOrders.containsKey(lower) || !this.acceptedMsg.containsKey(lower)) {
				this.myAgent.log("I do not yet received and accepted msg from :"+lower+" ", Logger.INFO_LEVEL);
				return null;
			}
			int cmp = this.bestOrder.compare(this.currentOrders.get(lower));
			int lowerRank = this.bestOrder.getOrder().indexOf(lower);
			if (cmp != 0 && Math.abs(cmp) <= lowerRank + 1) {
				this.myAgent.log("the rank of "+lower+" is not coherent with the best order", Logger.INFO_LEVEL);
				return null;
			}
			Instantiation view = this.acceptedMsg.get(lower);
			if (myView.compatible(view)) {
				myView.priorityMerge(view);
			} else {
				this.myAgent.log("The view of " + lower + " is not compatible :" + view, Logger.INFO_LEVEL);
				return null;
			}
		}
		for (int lower : lowerNeighbors) {
			Instantiation view = this.acceptedMsg.get(lower);
			if (myView.compareTimeCounters(view) != 0) {
				this.myAgent.log("The view of " + lower + " is not up to date :" + view, Logger.INFO_LEVEL);
				return null;
			}
		}
		return myView;
	}

	public void setBestOrder(CurrentOrder currentOrder) {
		if (this.bestOrder == null) {
			this.bestOrder = currentOrder;
		} else {
			if (this.bestOrder.compare(currentOrder) < 0) {
				this.bestOrder = currentOrder;
			}
		}
	}

	public void setCurrentOrder(CurrentOrder currentOrder) {
		this.currentOrders.put(this.myAgent.agentId, currentOrder);
		this.setBestOrder(currentOrder);
	}

}
