package simulation;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import dischoco.Protocols;

import kernel.AbstractAgent;
import kernel.AbstractMaster;
import simple.SimpleAgent;
import util.Commands;
import util.parser.InstanceParser;
import util.parser.PConstraint;
import util.parser.PVariable;

/**
 * A class that models a distributed problem. When a simulation is performed, this class encapsulates all informations for resolving distributed
 * problem: Master, Agents, variables and constraints. Also it allows to load problem from XML file. This class defines methods to create any type of
 * constraint.
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class DisProblem {
	/**
	 * The name of the problem to solve
	 */
	private String pbName;
	/**
	 * A map of the agents that are given by their IDs
	 */

	public HashMap<Integer, AbstractAgent> agents;
	/**
	 * The Master controlling the resolution of the problem
	 */
	public AbstractMaster master = null;

	InstanceParser parser;
	private String protocol;
	private String problemFile;

	/**
	 * Constructs a new problem instance
	 * 
	 * @param name
	 *            the problem name
	 * @param master
	 *            the Master controlling the resolution
	 */
	public DisProblem(String name, AbstractMaster master) {
		this.agents = new HashMap<Integer, AbstractAgent>();
		this.pbName = name;
		this.master = master;
	}

	/**
	 * Constructs a new problem instance
	 * 
	 * @param the
	 *            algorithm name to solve the instance;
	 * @param master
	 *            the Master controlling the resolution
	 */
	public DisProblem(InstanceParser parser, String protocol) {
		this.parser = parser;
		this.agents = new HashMap<Integer, AbstractAgent>();
		this.pbName = parser.getProblemName();
		this.master = Protocols.getMaster(protocol);
		this.loadProblem(parser);
	}

	public DisProblem(String problemFile, String protocol) {
		parser = new InstanceParser(problemFile);
		parser.loadInstance(problemFile);
		parser.parse(false);
		this.agents = new HashMap<Integer, AbstractAgent>();
		this.pbName = parser.getProblemName();
		this.master = Protocols.getMaster(protocol);
		this.setProtocol(protocol);
		this.loadProblem(parser);
		this.problemFile = problemFile;
	}

	/**
	 * make and add a new agent to the problem
	 */
	public AbstractAgent makeAgent(String agentName, String description) {
		if (this.getAgent(agentName) != null)
			throw new Error("An agent with the same name was already created");
		AbstractAgent agent = this.master.makeAgent(this.getNumAgents() + 1, agentName, description);
		this.agents.put(agent.getAgentId(), agent);
		return agent;
	}

	/**
	 * Returns the agent with the given name
	 * 
	 * @param agentName
	 *            the name of the agent to return
	 * @return
	 */
	public AbstractAgent getAgent(String agentName) {
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			if (ag.getAgentName().equals(agentName))
				return ag;
		}
		return null;
	}

	/**
	 * Returns the total number of Agents
	 */
	public int getNumAgents() {
		return this.agents.size();
	}

	/**
	 * Returns the map of agents by by their IDs
	 */
	public HashMap<Integer, AbstractAgent> getAgents() {
		return this.agents;
	}

	/**
	 * Returns the map of agents by by their IDs
	 */
	public Set<String> getAgentsNames() {
		Set<String> names = new TreeSet<String>();
		for(int id : this.agents.keySet()){
			names.add(this.agents.get(id).getAgentName());
		}
		return names;
	}

	/**
	 * Adds an new agent to the list of agents that participates in the resolution
	 */
	public void addAgent(AbstractAgent agent) {
		this.agents.put(agent.getAgentId(), agent);
	}

	/**
	 * Returns the agent with the given ID
	 * 
	 * @param agId
	 * 
	 */
	public AbstractAgent getAgent(int agId) {
		if (this.agents.containsKey(agId))
			return this.agents.get(agId);
		return null;
	}

	/**
	 * Returns the agent with the given ID
	 * 
	 * @param agId
	 * 
	 */
	public String getName(int agId) {
		if (this.agents.containsKey(agId))
			return this.agents.get(agId).getAgentName();
		return "A"+agId;
	}

	/**
	 * Set the list of the agents to the given one.
	 * 
	 * @param agents
	 */
	public void setAgents(List<? extends AbstractAgent> agents) {
		if (agents == null && this.agents != null) {
			this.agents.clear();
			this.agents = null;
		} else{
			for (AbstractAgent ag : agents) {
				this.addAgent(ag);
			}
		}
	}

	/**
	 * Solve the problem
	 */
	public void solve() {
		DisSolverSimulator disSolver = new DisSolverSimulator(this);
		disSolver.solve();
	}

	/**
	 * Returns a string representation of the problem
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer(this.pbName + ": \n");
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			sb.append(ag.toString() + "\n");
		}
		return sb.toString();
	}

	/**
	 * Loads Distributed Problem Instance from the given XML parser
	 * 
	 * @param parser
	 */

	public void loadProblem(InstanceParser parser) {
		if (parser.getModel().equals(Commands.COMPLEX) && this.master.getModel().equals(Commands.SIMPLE))
			parser.convertToSimpleModel();
		this.setAgents(master.makeAgents(parser));
		if (parser.getModel().equals(Commands.SIMPLE)) {
			this.buildSimpleProblem(parser);
			this.groupLocalAgents();
		} else{
			this.buildComplexProblem(parser);
		}
	}

	/**
	 * A distributed problem with complex local problem (agents with multiple variables) can be decomposed by combining for each local variable a
	 * pseudo-agent or an agent. In such a case, for not counting the messages sent between the pseudo-agents, agents must know its local neighbors.
	 * So, a call to this method allows agents to know its local neighbors
	 */
	private void groupLocalAgents() {
		Integer[] agIds = new Integer[this.agents.size()];
		agIds = agents.keySet().toArray(agIds);
		for (int i = 0 ; i < agIds.length ; i++) {
			AbstractAgent agent1 = agents.get(agIds[i]);
			for (int j = i + 1 ; j < agIds.length ; j++) {
				AbstractAgent agent2 = agents.get(agIds[j]);
				if (agent1.description.equals(agent2.description)) {
					agent2.addLocalAgent(agent1.getAgentId());
					agent1.addLocalAgent(agent2.getAgentId());
				}
			}
		}
	}

	/**
	 * to build a simple problem instance (simple: one variable per agent)
	 */
	private void buildSimpleProblem(InstanceParser parser) {
		Iterator<String> it = parser.getMapOfVariables().keySet().iterator();
		while (it.hasNext()) {
			PVariable pv = parser.getMapOfVariables().get(it.next());
			int agId = pv.getAgent().getId();
			SimpleAgent agent = (SimpleAgent) this.getAgent(agId);
			agent.var = pv.makeIntVariable(agent);
		}
		it = parser.getMapOfConstraints().keySet().iterator();
		while (it.hasNext()) {
			PConstraint constr = parser.getMapOfConstraints().get(it.next());
			constr.makeConstraints();
		}
	}

	/**
	 * to build a complex problem (Complex: Agent with multiple variable)
	 * 
	 * @param parser
	 */

	private void buildComplexProblem(InstanceParser parser) {
	// TODO
	}

	public String getProblemName() {
		return this.pbName;
	}

	public String getProblemFile() {
		return this.problemFile;
	}

	public void clear() {
		parser.clear();
		this.parser = null;

	}

	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}

	public String getProtocol() {
		return protocol;
	}
}
