package util;

import java.io.Serializable;
import java.util.StringTokenizer;

import kernel.VarIdentifier;

/**
 * A class models the variable assignment. A variable assignment is a data structure that assigns a value to the variable. It contains also the time
 * counter of the assignment
 * 
 * @author Redouane EZZAHIR
 * 
 */
public class VarAssignment implements Assignment, IVarIdentifier, Serializable {

	private static final long serialVersionUID = 2207542039642548070L;
	/**
	 * The assigned variable identifier
	 */
	protected VarIdentifier varId;
	/**
	 * The assigned value for the variable
	 */
	private int value = Integer.MIN_VALUE;
	/**
	 *A time counter that store the last time (Integer) of the assignment
	 */
	private int timeCounter = 0;

	/**
	 * Constructs a variable assignment without time counter
	 * 
	 * @param varId
	 *            the variable identifier
	 * @param value
	 *            the assigned value
	 */
	public VarAssignment(VarIdentifier varId, int value) {
		this.varId = varId;
		this.value = value;
	}

	/**
	 * Constructs a variable assignment with a time counter
	 * 
	 * @param varId
	 *            the variable identifier
	 * @param value
	 *            the assigned value
	 * @param timeCounter
	 *            the time counter of the assignment
	 */
	public VarAssignment(VarIdentifier varId, int value, int timeCounter) {
		this.varId = varId;
		this.value = value;
		this.timeCounter = timeCounter;
		this.hashCode();
	}

	/**
	 * Constructs a variable assignment for simple mode (Single variable per agent, thus index of variable is 0)
	 * 
	 * @param agId
	 *            the ID of the simple agent owning the variable
	 * @param value
	 *            the assigned value
	 * @param timeCounter
	 *            the time counter of the assignment
	 */
	public VarAssignment(int agId, int value, int timeCounter) {
		this(new VarIdentifier(agId, 0), value, timeCounter);
	}

	/**
	 * Constructs a variable assignment
	 * 
	 * @param agId
	 *            the ID of the simple agent owning the variable
	 * @param value
	 *            the assigned value
	 * @param timeCounter
	 *            the time counter of the assignment
	 */

	public VarAssignment(int agId, int varIndex, int value, int timeCounter) {
		this(new VarIdentifier(agId, varIndex), value);
		this.value = value;
		this.timeCounter = timeCounter;
	}

	/**
	 * Set the assigned value to the given one
	 * 
	 * @param val
	 *            the value assigned to the variable
	 */
	public void setValue(int val) {
		this.value = val;
	}

	/**
	 * Returns the value assigned to the variable
	 */
	public int getValue() {
		return this.value;
	}

	/**
	 * Return the value of the time counter
	 */
	public int getTimeCounter() {
		return this.timeCounter;
	}

	/**
	 * Set the value of the time counter to the given value
	 */

	public void setTimeCounter(int counter) {
		this.timeCounter = Math.max(this.timeCounter, counter);
	}

	/**
	 * Gets the ID of the agent owning (that had assign) the variable
	 */

	public int getAgentId() {
		return this.varId.getAgentId();
	}

	/**
	 * Returns the index of the variable
	 */

	public int getVarIndex() {
		return this.varId.getVarIndex();
	}

	/**
	 * Returns the identifier of the variable
	 */

	public VarIdentifier getVarIdentifier() {
		return varId;
	}

	/**
	 * Checks whether this variable assignment is equals to the given object or not.
	 */
	public boolean equals(Object o) {
		VarAssignment v = (VarAssignment) o;
		return this.varId.equals(v.varId) && this.value == v.value && this.timeCounter == v.timeCounter;

	}

	public int hashCode() {
		return this.varId.hashCode() + this.value + this.timeCounter;
	}

	/**
	 * Returns a clone of this object.
	 */

	public VarAssignment clone() {
		return new VarAssignment(this.varId.clone(), this.value, this.timeCounter);
	}

	public void setVarIndex(int index) {
		this.varId = new VarIdentifier(this.varId.getAgentId(), index);
	}

	/**
	 * 
	 * {@inheritDoc}
	 */
	public String toString() {
		return "(" + this.varId + "=" + this.value + "/" + this.timeCounter + ")";
	}

	/**
	 * Create a variable assignment from the given string
	 * 
	 * @param str
	 *            a string that represents the variable assignment.
	 */
	public static VarAssignment fromString(String str) {
		str = str.substring(1, str.indexOf(")"));
		String[] s = str.split("/");
		StringTokenizer t = new StringTokenizer(s[0], "=");
		char[] sa = t.nextToken().toCharArray();
		int i = 0;
		String agId = new String();
		String vId = new String();
		for ( ; i < sa.length ; i++) {
			if (sa[i] == '.')
				break;
			agId += String.valueOf(sa[i]);
		}
		i++;
		for ( ; i < sa.length ; i++) {
			vId += String.valueOf(sa[i]);
		}
		int agentId = new Integer(agId);
		int varIndex = new Integer(vId);
		VarIdentifier vvId = new VarIdentifier(agentId, varIndex);
		int val = Integer.parseInt(t.nextToken());
		int counter = Integer.parseInt(s[1]);
		return new VarAssignment(vvId, val, counter);
	}

	/**
	 * Compare the time counter of this variable assignment with these of the given one. Returns this.timeCounter - v.timeCounter
	 */
	public int compareTimeCounter(VarAssignment v) {
		if (v == null)
			return Integer.MAX_VALUE;
		if (this.varId.equals(v.varId)) { return this.timeCounter - v.timeCounter; }
		throw new Error("Try to compare time counter of two different variables : " + this + " and " + v);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.Assignment#compatible(util.Assignment)
	 */
	public boolean compatible(Assignment o) {
		if (o instanceof VarAssignment)
			return this.compatible((VarAssignment) o);
		else if (o instanceof Instantiation) { return ((Instantiation) o).compatible(this); }
		throw new Error("Compatibilty check error VarAssignment::compatible(" + o + ")");
	}

	/**
	 * Unassigns variable. It removes the value assigned to this variable.
	 */
	public void clear() {
		this.value = Integer.MIN_VALUE;
		this.timeCounter = 0;
	}

	/**
	 * Set this variable assignment to be private. It simply assigns the value of the timeCounter to the value this.value = this.timeCounter;
	 */
	public void setPrivate() {
		this.value = this.timeCounter;
	}

	/**
	 * Checks whether a value is assigned or not
	 * 
	 * @return
	 */

	public boolean isInstantiated() {
		return this.value != Integer.MIN_VALUE;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */

	public int compareTo(IVarIdentifier v) {
		if (v != null) {
			int d = this.varId.getAgentId() - v.getAgentId();
			if (d == 0)
				d = this.varId.getVarIndex() - v.getVarIndex();
			return d;
		}
		return Integer.MAX_VALUE;
	}

}
