package util.graph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Stack;
import java.util.StringTokenizer;

import util.Tools;

/**
 * An example class for directed graphs. The vertex type can be specified. There are no edge costs/weights.
 * 
 * Written for CS211, Nov 2006.
 * 
 * @author Paul Chew
 */
public class Digraph<V> {

	/**
	 * The implementation here is basically an adjacency list, but instead of an array of lists, a Map is used to map each vertex to its list of
	 * adjacent vertices.
	 */
	private Map<V, List<V>> graph = new HashMap<V, List<V>>();

	public int size() {
		return this.graph.size();
	}

	/**
	 * Returns a string representation of the Digraph
	 */
	public String toString() {
		StringBuffer s = new StringBuffer();
		for (V v : graph.keySet())
			s.append("\n    " + v + " -> " + graph.get(v));
		return s.toString();
	}

	/**
	 * Add a vertex to the graph. Nothing happens if vertex is already in graph.
	 */
	public void add(V vertex) {
		if (graph.containsKey(vertex))
			return;
		graph.put(vertex, new ArrayList<V>());
	}

	/**
	 * remove a vertex frm the graph. Nothing happens if vertex is not already in graph.
	 */
	public void remove(V vertex) {
		if (!graph.containsKey(vertex))
			return;
		graph.remove(vertex);
		for (V e : graph.keySet()) {
			if (graph.get(e).contains(vertex))
				graph.get(e).remove(vertex);
		}
	}

	/**
	 * True iff graph contains vertex.
	 */
	public boolean contains(V vertex) {
		return graph.containsKey(vertex);
	}

	/**
	 * Add an edge to the graph; if either vertex does not exist, it's added. This implementation allows the creation of multi-edges and self-loops.
	 */
	public void add(V from, V to) {
		this.add(from);
		this.add(to);
		if (!graph.get(from).contains(to))
			graph.get(from).add(to);
	}

	/**
	 * Remove an edge from the graph. Nothing happens if no such edge.
	 * 
	 * @throws IllegalArgumentException
	 *             if either vertex doesn't exist.
	 */
	public void remove(V from, V to) {
		if (!(this.contains(from) && this.contains(to)))
			throw new IllegalArgumentException("Nonexistent vertex");
		graph.get(from).remove(to);
	}

	/**
	 * Report (as List) the graph of the gevin vertex
	 */
	public List<V> getNeighbors(V from) {
		if (this.contains(from))
			return this.graph.get(from);
		return new ArrayList<V>();
	}

	/**
	 * Report (as a Map) the out-degree of each vertex.
	 */
	public Map<V, Integer> outDegree() {
		Map<V, Integer> result = new HashMap<V, Integer>();
		for (V v : graph.keySet())
			result.put(v, graph.get(v).size());
		return result;
	}

	/**
	 * Report (as a List) the set of vertex that must follow the given vertex.
	 */
	public List<V> mustFollow(V vertex) {
		List<V> result = new ArrayList<V>();
		if (this.contains(vertex)) {
			result = follow(vertex);
			for (V w : graph.get(vertex))
				result.remove(w);
		}
		return result;
	}

	private List<V> follow(V vertex) {
		List<V> result = new ArrayList<V>();
		for (V w : graph.get(vertex)) {
			if (!result.contains(w))
				result.add(w);
			List<V> result2 = follow(w);
			for (V z : result2)
				if (!result.contains(z))
					result.add(z);
		}
		return result;
	}

	public int compare(V x, V y) {
		if (this.contains(x) && this.contains(y)) {
			if (mustFollow(x).contains(y))
				return 1;
			if (mustFollow(y).contains(x))
				return -1;
		}
		return 0;
	}

	/**
	 * If this Directed graph is used to represent a partial order the weakening at x of the partial order try to remove from the graph all safety
	 * conditions z < y where x < y and than adding the safety condition x < y .
	 * 
	 */
	public void weakening(V x) {
		List<V> mfx = mustFollow(x);
		for (V y : mfx) {
			for (V v : graph.keySet()) {
				if (!x.equals(v))
					if (graph.get(v).contains(y))
						this.remove(v, y);
			}
			this.add(x, y);
		}

	}

	/**
	 * Report (as a Map) the in-degree of each vertex.
	 */
	public Map<V, Integer> inDegree() {
		Map<V, Integer> result = new HashMap<V, Integer>();
		for (V v : graph.keySet())
			result.put(v, 0); // All in-degrees are 0
		for (V from : graph.keySet()) {
			for (V to : graph.get(from)) {
				result.put(to, result.get(to) + 1); // Increment in-degree
			}
		}
		return result;
	}

	/**
	 * Report (as a List) the topological sort of the vertices; null for no such sort.
	 */
	public List<V> topologicalSort() {
		Map<V, Integer> degree = inDegree();
		// Determine all vertices with zero in-degree
		Stack<V> zeroVerts = new Stack<V>(); // Stack as good as any here
		for (V v : degree.keySet()) {
			if (degree.get(v) == 0)
				zeroVerts.push(v);
		}

		// Determine the topological order
		List<V> result = new ArrayList<V>();
		while (!zeroVerts.isEmpty()) {
			V v = zeroVerts.pop(); // Choose a vertex with zero in-degree
			result.add(v); // Vertex v is next in topol order
			// "Remove" vertex v by updating its graph
			for (V neighbor : graph.get(v)) {
				degree.put(neighbor, degree.get(neighbor) - 1);
				// Remember any vertices that now have zero in-degree
				if (degree.get(neighbor) == 0)
					zeroVerts.push(neighbor);
			}
		}
		// Check that we have used the entire graph (if not, there was a cycle)
		if (result.size() != graph.size())
			return null;
		return result;
	}

	/**
	 * Report (as a List) the topological sort of the vertices; null for no such sort. This version is based on the use of Depth First Search.
	 */
	public List<V> topologicalSortDFS() {
		List<V> result = new LinkedList<V>(); // Used to store topological sort
		Map<V, String> marked = new HashMap<V, String>(); // Map: vertex -> mark
		try {
			for (V v : graph.keySet()) {
				if (marked.containsKey(v))
					continue;
				dfs(v, marked, result);
			}
		} catch (RuntimeException e) {
			return null;
		}
		return result;
	}

	/**
	 * Private dfs for doing topSort2.
	 */
	private void dfs(V v, Map<V, String> marked, List<V> result) {
		marked.put(v, "active");
		for (V w : graph.get(v)) {
			if (!marked.containsKey(w))
				dfs(w, marked, result);
			if (marked.get(w) == "active")
				throw new RuntimeException();
		}
		marked.put(v, "done");
		result.add(0, v);
	}

	/**
	 * True iff graph is a dag (directed acyclic graph).
	 */
	public boolean isDag() {
		return topologicalSort() != null;
	}

	public boolean isCyclic() {
		return !isDag();
	}

	/**
	 * Report (as a Map) the bfs distance to each vertex from the start vertex. The distance is an Integer; the value null is used to represent
	 * infinity (implying that the corresponding node cannot be reached).
	 */
	public Map<V, Integer> bfsDistance(V start) {
		Map<V, Integer> distance = new HashMap<V, Integer>();
		// Initially, all distance are infinity, except start node
		for (V v : graph.keySet())
			distance.put(v, null);
		distance.put(start, 0);
		// Process nodes in queue order
		Queue<V> queue = new LinkedList<V>();
		queue.offer(start); // Place start node in queue
		while (!queue.isEmpty()) {
			V v = queue.remove();
			int vDist = distance.get(v);
			// Update graph
			for (V neighbor : graph.get(v)) {
				if (distance.get(neighbor) != null)
					continue; // Ignore if already done
				distance.put(neighbor, vDist + 1);
				queue.offer(neighbor);
			}
		}
		return distance;
	}

	/**
	 * Warning!: If you try to use a generic object is different to basic types:Integer, Double ...
	 */
	public Digraph<V> clone() {
		Digraph<V> dg = new Digraph<V>();
		for (V from : graph.keySet()) {
			for (V to : graph.get(from)) {
				dg.add(from, to);
			}
		}
		return dg;
	}

	/**
	 * Main program (for testing).
	 */

	public static Digraph<Integer> POrderFromString(String str) {
		Digraph<Integer> po = new Digraph<Integer>();
		if (str.equals(new String()) || str.equals("{}"))
			return po;
		str = str.substring(1);
		str = str.substring(0, str.indexOf('}') - 1);

		String[] strs = str.split("],");
		for (int i = 0 ; i < strs.length ; i++) {
			StringTokenizer tok = new StringTokenizer(strs[i], "=[");
			int from = Integer.parseInt(tok.nextToken().trim());
			int[] tos = null;
			if (tok.hasMoreElements()) {
				tos = Tools.IntegerArrayFromString(tok.nextToken(), ",");
				for (int j = 0 ; j < tos.length ; j++)
					po.add(from, tos[j]);
			} else
				po.add(from);
		}
		return po;
	}

	public static void main(String[] args) {
		// Create a Graph with Integer nodes
		Digraph<Integer> graph = new Digraph<Integer>();
		graph.add(0, 1);
		graph.add(0, 2);
		graph.add(0, 3);
		graph.add(1, 2);
		graph.add(1, 3);
		graph.add(2, 3);
		graph.add(2, 4);
		graph.add(4, 5);
		graph.add(5, 6); // Tetrahedron with tail
		System.out.println("The current graph: " + graph.toString());
		System.out.println("In-degrees: " + graph.inDegree());
		System.out.println("Out-degrees: " + graph.outDegree());
		System.out.println("A topological sort of the vertices: " + graph.topologicalSort());
		System.out.println("A topological sort of the vertices: " + graph.topologicalSortDFS());
		System.out.println("The graph " + (graph.isDag()? "is" : "is not") + " a dag");
		System.out.println("BFS distances starting from " + 0 + ": " + graph.bfsDistance(0));
		System.out.println("BFS distances starting from " + 1 + ": " + graph.bfsDistance(1));
		System.out.println("BFS distances starting from " + 2 + ": " + graph.bfsDistance(2));
		System.out.println("must follow : " + 0 + " : " + graph.mustFollow(0));
		Digraph<Integer> g = Digraph.POrderFromString("{2=[8], 9=[8], 8=[], 1=[8], 7=[8]}").clone();
		g.weakening(0);
		System.out.println("The weakening of the cloned graph at 0: " + g);
		graph.add(4, 1); // Create a cycle
		System.out.println("Cycle created");
		System.out.println("The current graph: " + graph.toString());
		System.out.println("A topological sort of the vertices: " + graph.topologicalSort());
		System.out.println("A topological sort of the vertices: " + graph.topologicalSortDFS());
		System.out.println("The graph " + (graph.isDag()? "is" : "is not") + " a dag");
		System.out.println("BFS distances starting from " + 2 + ": " + graph.bfsDistance(2));
	}

}
