package util.graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

public class Graph<V> {

	// symbol table: key = V vertex, value = set of neighboring vertices
	private Map<V, List<V>> graph = new HashMap<V, List<V>>();

	// number of edges
	private int numEdges;

	private Vector<V> explored;

	// create an empty graph
	public Graph() {}

	// add w to v's set of neighbors, and add v to w's set of neighbors
	public void addEdge(V v, V w) {
		if (!hasEdge(v, w)) {
			numEdges++;
			if (!hasVertex(v)) addVertex(v);
			if (!hasVertex(w)) addVertex(w);
			graph.get(v).add(w);
			graph.get(w).add(v);
		}
	}

	/**
	 * remove the graph of a vertex to the graph. Nothing happens if vertex is not already in graph.
	 */
	public void remove(V vertex) {
		if (!graph.containsKey(vertex)) return;
		Iterable<V> neig = this.getNeighbors(vertex);
		graph.remove(vertex);
		for (V e : neig) {
			numEdges--;
			if (graph.get(e).contains(vertex)) graph.get(e).remove(vertex);
		}
	}

	/**
	 * This method inserts a vertex into a graph. It add a new vertex v with no neighbors (if vertex does not yet exist)
	 * 
	 * @param v
	 */
	public void addVertex(V v) {
		if (!hasVertex(v)) graph.put(v, new ArrayList<V>());
	}

	/**
	 * This method returns the number of edges contained by the graph.
	 */
	public int getNumberOfEdges() {
		return numEdges;
	}

	/**
	 * This method returns the number of vertices contained by the graph.
	 */
	public int getNumberOfVertices() {
		return graph.size();
	}

	/**
	 * return the degree of vertex v
	 */
	public int degree(V v) {
		if (!this.graph.containsKey(v)) {
			return 0;
		} else {
			return graph.get(v).size();
		}
	}

	public Iterable<V> getVertices() {
		return graph.keySet();
	}

	// return an iterator over the neighbors of vertex v
	public Iterable<V> getNeighbors(V v) {
		// return empty set if vertex isn't in graph
		if (!hasVertex(v)) return new ArrayList<V>();
		else return graph.get(v);
	}

	/**
	 * is v a vertex in the graph?
	 * 
	 * @param v
	 */
	public boolean hasVertex(V v) {
		return graph.containsKey(v);
	}

	/**
	 * is v-w an edge in the graph?
	 * 
	 * @param v
	 * @param w
	 */
	public boolean hasEdge(V v, V w) {
		if (!hasVertex(v)) return false;
		return graph.get(v).contains(w);
	}

	public int maxDistance(V v) {
		int cnt = 0;
		int n = this.getNumberOfVertices();
		Map<V, Integer> dist = new HashMap<V, Integer>();
		for (V w : this.getVertices())
			dist.put(w, n); // set to maximum distance
		dist.put(v, 0);
		int depth = 1; // next distance in BFS
		ArrayList<V> distList = new ArrayList<V>();
		distList.add(v);
		while (distList.size() > 0) {
			ArrayList<V> nextList = new ArrayList<V>();
			for (V u : distList) {
				Iterable<V> nbrs = this.getNeighbors(u);
				for (V w : nbrs) {
					if (dist.get(w) == n) // first time reachable?
					{
						cnt++;
						dist.put(w, depth);
						nextList.add(u);
					}
				}
			}
			distList = nextList; // next try set of vertices further away
			depth++;
		}
		return cnt == n? depth - 1 : n;
	}

	public int diameter() {
		int max = 0;
		for (V v : this.getVertices()) {
			int d = maxDistance(v);
			max = (d > max? d : max);
		}
		return max;
	}

	/**
	 * This boolean-valued method returns true if the graph is cyclic.
	 */

	public boolean isCyclic() {
		if (this.getNumberOfEdges() >= this.getNumberOfVertices()) return true;
		Map<V, Integer> PreOrder = new HashMap<V, Integer>();
		Map<V, Integer> PostOrder = new HashMap<V, Integer>();
		for (V v : this.getVertices()) {
			PreOrder.put(v, 0);
			PostOrder.put(v, 0);
		}

		countPair cnt = new countPair();
		int components = 0;
		// 
		for (V v : this.getVertices()) {
			if (PreOrder.get(v) > 0) continue;
			// else try next component with root i
			// 
			this.doDFS(v, PreOrder, PostOrder, cnt);
			components++;
		}
		return this.getNumberOfEdges() != this.getNumberOfVertices() + components;

	}

	static private class countPair { // private counters for DFS search
		int cnt1, cnt2;
		int inc1() {
			return ++cnt1;
		} // increment and return counters

		int inc2() {
			return ++cnt2;
		}
	}

	public void doDFS(V v, Map<V, Integer> PreOrder, Map<V, Integer> PostOrder, countPair cnt) {
		PreOrder.put(v, cnt.inc1());
		for (V u : this.getNeighbors(v)) {
			if (PreOrder.get(u) == 0) { // Have we not seen vertex u before?
				doDFS(u, PreOrder, PostOrder, cnt);
			}
		}
		PostOrder.put(v, cnt.inc2());
		return;
	}

	/**
	 * An undirected graph G:(V,E) is connected if there is a path in G between every pair of vertices in V. This
	 * boolean-valued method returns true
	 * if the graph is connected.
	 * 
	 * @return true if the graph is connected.
	 */
	public boolean isConnected() {
		return BFS(this.graph.keySet().iterator().next()) == this.getNumberOfVertices();
	}

	public int BFS(V v) {
		Map<V, Integer> LevelOrder = new HashMap<V, Integer>();
		for (V u : this.getVertices()) {
			LevelOrder.put(u, 0);
		}
		LinkedList<V> toGrow = new LinkedList<V>();
		int cnt = 0;
		LevelOrder.put(v, ++cnt);
		toGrow.addLast(v);
		while (!toGrow.isEmpty()) {
			V grow = toGrow.removeFirst();
			Iterable<V> nbrs = this.getNeighbors(grow);
			for (V u : nbrs) {
				if (LevelOrder.get(u) == 0) {// i.e., not visited yet
					LevelOrder.put(u, ++cnt);
					toGrow.addLast(u);
				}
			}
		}
		return cnt;
	}

	/**
	 * Returns a string representation of the graph
	 * 
	 * @return
	 */
	public static String toMsg(Graph<Integer> G) {
		StringBuffer s = new StringBuffer();
		for (Integer v : G.getVertices()) {
			s.append(v + ":");
			for (Integer w : G.getNeighbors(v)) {
				s.append(w + " ");
			}
			s.append(";");
		}
		return s.toString();
	}

	public static Graph<Integer> fromMsg(String msg) {
		Graph<Integer> G = new Graph<Integer>();
		StringTokenizer tok = new StringTokenizer(msg, ";");
		while (tok.hasMoreTokens()) {
			String ver = tok.nextToken();
			StringTokenizer t = new StringTokenizer(ver, ":");
			if (t.hasMoreTokens()) {
				int v = Integer.parseInt(t.nextToken());
				if (t.hasMoreTokens()) {
					StringTokenizer e = new StringTokenizer(t.nextToken(), " ");
					while (e.hasMoreTokens()) {
						int w = Integer.parseInt(e.nextToken());
						G.addEdge(v, w);
					}
				}
			}
		}
		return G;
	}

	/**
	 * Returns a string representation of the graph
	 */
	public String toString() {
		StringBuffer s = new StringBuffer();
		for (V v : graph.keySet())
			s.append("\n    " + v + " -> " + graph.get(v));
		return s.toString();
	}
	
	/*
	public String toDFSString() {
		StringBuffer sb = new StringBuffer();
		sb.append(root.Id);
		for (PriorityTree st : children) {
			sb.append(" " + st.toDFSString());
		}
		return sb.toString();
	}
	
	 */

	/**
	 * @param findNextPriority
	 * @return
	 */
	public PriorityTree buildDFS(V v) {
		this.explored = new Vector<V>();
		PriorityTree tree = DFS(v);
		Collection<V> vv = this.graph.keySet();
		vv.removeAll(this.explored);
		if(vv.size()>0){
			tree.addSubTree(DFS(vv.iterator().next()));
		}
		return tree;
	}
	/**
	 * 
	 */
	public PriorityTree DFS(V v){
		PriorityTree tree = new PriorityTree(new Integer(v+""));
		explore(v);
//		System.out.println(v+"--");
		for (V u : this.getNeighbors(v)){
			if(!isExplored(u)){
				tree.addSubTree(DFS(u));
			}
		}
		return tree ;
	}
	
	/**
	 * 
	 */
	public PriorityTree buildMaxDynDegDFS(V v){
		this.explored = new Vector<V>();
		return this.MaxDynDegDFS(v);
	}
	/**
	 * 
	 */
	public PriorityTree MaxDynDegDFS(V v){
		PriorityTree tree = new PriorityTree(new Integer(v+""));
		explore(v);
		while(this.getUnexploredNeighbors(v).size()>0){
			V u = this.selectNeighbor(this.getUnexploredNeighbors(v));
			tree.addSubTree(MaxDynDegDFS(u));
		}
		return tree ;
	}

	/**
	 * @param vector
	 * @return
	 */
	private V selectNeighbor(Vector<V> vector) {
		HashMap<V, Integer> map = new HashMap<V, Integer>();
		for(V u : vector){
			map.put(u, this.getUnexploredNeighbors(u).size());
		}
		int max =-1;
		V selected = null;
		for(V u : map.keySet()){
			if(map.get(u) > max){
				selected = u;
				max = map.get(u) ;
			}
		}
		return selected;
	}

	/**
	 * @param u
	 * @return
	 */
	private Vector<V> getUnexploredNeighbors(V v) {
		Vector<V> ans = new Vector<V>();
		for(V u : this.getNeighbors(v)){
			if(!this.explored.contains(u)){
				ans.add(u);
			}
		}
		return ans;
	}

//	/**
//	 * @param v
//	 * @return
//	 */
//	private boolean hasUnexploredNeighbor(V v) {
//		for(V u : this.getNeighbors(v)){
//			if(!this.explored.contains(u)){
//				return true;
//			}
//		}
//		return false;
//	}
	/**
	 * @param v
	 */
	private void explore(V v) {
		this.explored.add(v);
	}

	/**
	 * @param v
	 * @return
	 */
	private boolean isExplored(V v) {
		return this.explored.contains(v);
	}

	public static void main(String[] args) {
		Graph<String> G = new Graph<String>();
		G.addEdge("A", "B");
		G.addEdge("A", "C");
		G.addEdge("C", "D");
		G.addEdge("D", "E");
		G.addEdge("D", "G");
		G.addEdge("E", "G");
		G.addVertex("H");

		// print out graph
		System.out.println(G.toString() + "\n G is cyclic? : " + G.isCyclic() + "\n G is connected? :"
				+ G.isConnected());

		// print out graph again by iterating over vertices and edges
		for (String v : G.getVertices()) {
			System.out.print(v + ": ");
			for (String w : G.getNeighbors(v)) {
				System.out.print(w + " ");
			}
			System.out.println();
		}

		Graph<Integer> g = new Graph<Integer>();
		g.addEdge(1, 2);
		g.addEdge(1, 4);
		g.addEdge(1, 7);
		g.addEdge(2, 7);
		g.addEdge(3, 4);
		g.addEdge(3, 5);
		g.addEdge(4, 5);
		g.addEdge(5, 7);
		System.out.println(g.toString() + "\n" + Graph.fromMsg(Graph.toMsg(g)) + "\n " + "is connected? :"
				+ g.isConnected());
	}
}
